#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## VII Coupler
# Paul Dechamps

from dart.viscous.Master.DAirfoil import Airfoil
from dart.viscous.Master.DWake import Wake
from dart.viscous.Drivers.DGroupConstructor import GroupConstructor

from fwk.coloring import ccolors

import fwk
import numpy as np
from matplotlib import pyplot as plt

class Coupler:
    def __init__(self, iSolver, vSolver) -> None:
        self.iSolver=iSolver
        self.vSolver=vSolver

        self.maxCouplIter = 100
        self.couplTol = 1e-4

        self.tms=fwk.Timers()
        pass

    def CreateProblem(self, _boundaryAirfoil, _boundaryWake):

        self.group = [Airfoil(_boundaryAirfoil), Wake(_boundaryWake)] # airfoil and wake python objects
        self.blwVel = [None, None, None]
        self.deltaStar = [None, None, None]
        self.xx = [None, None, None]
        pass

    def Run(self):

      # Initilize meshes in c++ objects

      nElmUpperPrev = 0
      couplIter = 0
      cdPrev = 0
      while couplIter < self.maxCouplIter:
        

        # Run inviscid solver
        self.tms['inv'].start()
        self.iSolver.run()
        self.tms['inv'].stop()

        # Extract inviscid boundary

        for n in range(0, len(self.group)):

          for i in range(0, len(self.group[n].boundary.nodes)):

            self.group[n].v[i,0] = self.iSolver.U[self.group[n].boundary.nodes[i].row][0]
            self.group[n].v[i,1] = self.iSolver.U[self.group[n].boundary.nodes[i].row][1]
            self.group[n].v[i,2] = self.iSolver.U[self.group[n].boundary.nodes[i].row][2]
            self.group[n].Me[i] = self.iSolver.M[self.group[n].boundary.nodes[i].row]
            self.group[n].rhoe[i] = self.iSolver.rho[self.group[n].boundary.nodes[i].row]
        
        dataIn=[None,None]
        for n in range(0,len(self.group)):
          self.group[n].connectListNodes, self.group[n].connectListElems, dataIn[n]  = self.group[n].connectList()
        fMeshDict, cMeshDict, data = GroupConstructor().mergeVectors(dataIn, 0, 1)

        if ((len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])) != (nElmUpperPrev or couplIter==0)):
          
          # Initialize mesh          
          self.vSolver.InitMeshes(0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]], fMeshDict['globCoord'][:fMeshDict['bNodes'][1]])
          self.vSolver.InitMeshes(1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['globCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
          self.vSolver.InitMeshes(2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:], fMeshDict['globCoord'][fMeshDict['bNodes'][2]:])

          self.vSolver.SendExtVariables(0, fMeshDict['locCoord'][:fMeshDict['bNodes'][1]], fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
          self.vSolver.SendExtVariables(1, fMeshDict['locCoord'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])    
          self.vSolver.SendExtVariables(2, fMeshDict['locCoord'][fMeshDict['bNodes'][2]:], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])
       
        else:

          self.vSolver.SendExtVariables(0, fMeshDict['xxExt'][:fMeshDict['bNodes'][1]], fMeshDict['deltaStarExt'][:fMeshDict['bNodes'][1]])
          self.vSolver.SendExtVariables(1, fMeshDict['xxExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
          self.vSolver.SendExtVariables(2, fMeshDict['xxExt'][fMeshDict['bNodes'][2]:], fMeshDict['deltaStarExt'][fMeshDict['bNodes'][2]:])

        nElmUpperPrev = len(fMeshDict['locCoord'][:fMeshDict['bNodes'][1]])
        self.vSolver.SendInvBC(0, fMeshDict['vtExt'][:fMeshDict['bNodes'][1]], fMeshDict['Me'][:fMeshDict['bNodes'][1]], fMeshDict['rho'][:fMeshDict['bNodes'][1]])
        self.vSolver.SendInvBC(1, fMeshDict['vtExt'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['Me'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]], fMeshDict['rho'][fMeshDict['bNodes'][1]:fMeshDict['bNodes'][2]])
        self.vSolver.SendInvBC(2, fMeshDict['vtExt'][fMeshDict['bNodes'][2]:], fMeshDict['Me'][fMeshDict['bNodes'][2]:], fMeshDict['rho'][fMeshDict['bNodes'][2]:])
        print('Inviscid boundary imposed')
        # Run viscous solver

        self.tms['visc'].start()
        exitCode = self.vSolver.Run(couplIter)
        self.tms['visc'].stop()
        print('Viscous ops terminated with exit code ', exitCode)

        error = abs((self.vSolver.Cdt - cdPrev)/self.vSolver.Cdt) if self.vSolver.Cdt != 0 else 1

        if error  <= self.couplTol and exitCode==0:

          print(self.tms)
          print('')
          print(ccolors.ANSI_GREEN, 'It: {:<3.0f} Cd = {:<6.5f}. Error = {:<2.3f} < {:<2.3f}'
          .format(couplIter, self.vSolver.Cdt, np.log10(error), np.log10(self.couplTol)), ccolors.ANSI_RESET)
          print('')
          return 0
        
        cdPrev = self.vSolver.Cdt
        
        print('')
        print(ccolors.ANSI_BLUE, 'It: {:<3.0f} Cd = {:<6.5f}. Error = {:<2.3f} > {:<2.3f}'
        .format(couplIter, self.vSolver.Cdt, np.log10(error), np.log10(self.couplTol)), ccolors.ANSI_RESET)
        print('')

        # Retreive and set blowing velocities.

        for iRegion in range(3):
          self.blwVel[iRegion] = np.asarray(self.vSolver.ExtractBlowingVel(iRegion))
          self.deltaStar[iRegion] = np.asarray(self.vSolver.ExtractDeltaStar(iRegion))
          self.xx[iRegion] = np.asarray(self.vSolver.ExtractXx(iRegion))

        blwVelAF = np.concatenate((self.blwVel[0], self.blwVel[1][1:])) # Remove stagnation point doublon.
        deltaStarAF = np.concatenate((self.deltaStar[0], self.deltaStar[1][1:])) # Remove stagnation point doublon.
        xxAF = np.concatenate((self.xx[0], self.xx[1][1:])) # Remove stagnation point doublon.
        blwVelWK = self.blwVel[2]
        deltaStarWK = self.deltaStar[2]
        xxWK = self.xx[2]

        self.group[0].u = blwVelAF[self.group[0].connectListElems.argsort()]
        self.group[1].u = blwVelWK[self.group[1].connectListElems.argsort()]

        self.group[0].deltaStar = deltaStarAF[self.group[0].connectListNodes.argsort()]
        self.group[1].deltaStar = deltaStarWK[self.group[1].connectListNodes.argsort()]

        self.group[0].xx = xxAF[self.group[0].connectListNodes.argsort()]
        self.group[1].xx = xxWK[self.group[1].connectListNodes.argsort()]

        for n in range(0, len(self.group)):
          for i in range(0, self.group[n].nE):
            self.group[n].boundary.setU(i, self.group[n].u[i])

        couplIter += 1
      else:
        print(ccolors.ANSI_RED, 'Maximum number of {:<.0f} coupling iterations reached'.format(self.maxCouplIter), ccolors.ANSI_RESET)
        return couplIter

