################################################################################
#                                                                              #
#                            DARTFlo viscous module file                       #
#                Driver : Communication with coupler and initialization        #
#                 of some components of the viscous computation part           #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
from dart.viscous.Drivers.DGroupConstructor2 import GroupConstructor
from dart.viscous.Drivers.DPostProcess import PostProcess

from dart.viscous.Solvers.DIntegration import Integration
from dart.viscous.Solvers.DSolver import Solver as TSolver
from dart.viscous.Solvers.DSysMatrix import sysMatrix
from dart.viscous.Solvers.DTransition import Transition

from dart.viscous.Physics.DVariables import Variables
from dart.viscous.Physics.DDiscretization import Discretization
from dart.viscous.Physics.DValidation import Validation
import dart.viscous.Physics.DBIConditions as Conditions

from fwk.coloring import ccolors

import numpy as np
import matplotlib.pyplot as plt

class Driver:
    """Driver of the time-dependent boundary layer equation solver (viscous solver)
    """

    def __init__(self,_Re, user_xtr, _CFL0, _Minfty, _case, _mapMesh, _nFactor):
        self.Re          = _Re
        self.Minfty      = _Minfty

        self.xtrF        = user_xtr
        self.Ti          = None
        self.mapMesh     = _mapMesh
        self.nFactor     = _nFactor
        self.xtr         = [1,1]
        self.CFL0        = _CFL0
        self.case        = _case

        self.Cd          = 0
        self.it          = 0
        self.uPrevious   = None
        self.nbrElmUpper = -1
        self.stagPtMvmt  = 0
        pass

    def dictionary(self):
        """Create dictionary with the coordinates and the boundary layer parameters
        """

        wData = {}
        wData['x']          = self.data[:,0]
        wData['y']          = self.data[:,1]
        wData['z']          = self.data[:,2]
        wData['x/c']        = self.data[:,0]/max(self.data[:,0])
        wData['Cd']         = self.blScalars[0]
        wData['Cdf']        = self.blScalars[1]
        wData['Cdp']        = self.blScalars[2]
        wData['delta*']     = self.blVectors[0]
        wData['theta']      = self.blVectors[1]
        wData['H']          = self.blVectors[2]
        wData['Hk']         = self.blVectors[3]
        wData['H*']         = self.blVectors[4]
        wData['H**']        = self.blVectors[5]
        wData['Cf']         = self.blVectors[6]
        wData['Cdissip']    = self.blVectors[7]
        wData['Ctau']       = self.blVectors[8]
        wData['delta']      = self.blVectors[9]
        if self.xtrF[0] is not None:
            wData['xtrTop'] = self.xtrF[0]
        else:
            wData['xtrTop'] = self.xtr[0]
        if self.xtrF[1] is not None:
            wData['xtrBot'] = self.xtrF[1]
        else:
            wData['xtrBot'] = self.xtr[1]
        self.wData = wData
        return wData

    def writeFile(self):
        """Write the results in airfoil_viscous.dat
        """
        
        wData = self.dictionary()
        toW = ['delta*', 'H', 'Cf', 'Ctau'] # list of variable to be written (should be a user input)
        # Write
        print('Writing file: airfoil_viscous.dat...', end = '')
        f = open('airfoil_viscous.dat', 'w+')

        f.write('$Aerodynamic coefficients\n')
        f.write('             Re              Cd             Cdp             Cdf         xtr_top         xtr_bot\n')
        f.write('{0:15.6f} {1:15.6f} {2:15.6f} {3:15.6f} {4:15.6f} {5:15.6f}\n'.format(self.Re, wData['Cd'], wData['Cdp'],
                                                                                        wData['Cdf'], wData['xtrTop'],
                                                                                        wData['xtrBot']))
        f.write('$Boundary layer variables\n')
        f.write('              x               y               z             x/c')

        for s in toW:
            f.write(' {0:>15s}'.format(s))
        f.write('\n')

        for i in range(len(self.data[:,0])):
            f.write('{0:15.6f} {1:15.6f} {2:15.6f} {3:15.6f}'.format(wData['x'][i], wData['y'][i], wData['z'][i], wData['x/c'][i]))
            for s in toW:
                f.write(' {0:15.6f}'.format(wData[s][i]))
            f.write('\n')

        f.close()
        print('done.')


    def run(self, groups):
        """Runs viscous solver driver.
        - Data preprocessing and initialization of the different components.
        - Runs the pseudo-time solver
        - Data postprocessing to ensure proper communication between the solvers.
        
        Args
        ----
        
        - groups (data structure): Structures data containing the necessary information
        related to each group (Airfoil upper surface, airfoil lower surface and wake).
        """

        nFactor = self.nFactor

        # Initialize computation structures.

        dataIn = [None,None]  # Merge upper, lower sides and wake
        for n in range(0, len(groups)):
          groups[n].connectListNodes, groups[n].connectListElems, dataIn[n]  = groups[n].connectList()
        fMeshDict, cMeshDict, data = GroupConstructor().mergeVectors(dataIn, self.mapMesh, nFactor)

        var = Variables(fMeshDict, self.xtrF, self.Ti, self.Re)                 # Initialize data container for flow config.

        if self.it != 0:
          var.extPar[4::var.nExtPar] = fMeshDict['xxExt']       	            	# Extenal flow local markers.

        DirichletBC=Conditions.Boundaries(self.Re)                              # Boundary condition (Airfoil/Wake).

        gradComp=Discretization(var.xx, var.extPar[4::var.nExtPar], var.bNodes) # Gradients computation.
        
        sysSolver = sysMatrix(var.nN, var.nVar, gradComp.fou,gradComp.fou)			# Boundary layer equations solver that can provide
                                                                                # Residuals and Jacobian computation.

        transSolver = Transition(var.xtrF)                                      # Transition solver.

        # Initialize time Integrator

        tIntegrator = Integration(sysSolver, self.Minfty, self.CFL0)
        tSolver=TSolver(tIntegrator, transSolver, DirichletBC.wakeBC)

        # Output setup at the first iteration.

        if self.it == 0:
            print('+------------------------------- Solver setup ---------------------------------+')
            print('- Reynolds number: {:>1.2e}'.format(self.Re))
            if self.mapMesh:
                print('- Mesh mapped from {:>5.0f} to {:>5.0f} points.'.format(len(cMeshDict['locCoord']), var.nN))
            print('- Number of points: {:>5.0f}.'.format(var.nN))
            if var.xtrF[0] is None:
                print('- Free transition on the upper side.')
            else:
                print('- Forced transition on the upper side; x/c = {:<.4f}.'.format(var.xtrF[0]))
            if var.xtrF[1] is None:
                print('- Free transition on the lower side.')
            else:
                print('- Forced transition on the lower side; x/c = {:<.4f}.'.format(var.xtrF[1]))
            print('- Critical amplification ratio: {:<.2f}.'.format(var.Ncrit))
            if var.xtrF == [0, 0]:
              print('- Turbulent stagnation point. Ct = 0.03')

            print('')
            print('Numerical parameters')
            print('- CFL0: {:<3.2f}'.format(tSolver.integrator.GetCFL0()))
            print('- Tolerance: {:<3.0f}'.format(np.log10(tSolver.integrator.Gettol())))
            print('- Solver maximum iterations: {:<5.0f}'.format(tSolver.integrator.GetmaxIt()))
            print('+------------------------------------------------------------------------------+')

        # Output preprocessing satut.

        print('- Mach max: {:<.2f}.'.format(np.max((var.extPar[0::var.nExtPar]))), end = ' ')

        # Solver setup.

        if self.uPrevious is None or self.nbrElmUpper != var.bNodes[1] or self.stagPtMvmt:

          tSolver.integrator.itFrozenJac0 = 1
          tSolver.initSln = 1   # Flag to use time solver initial condition.
          tSolver.integrator.SetCFL0(1) # Low starting CFL.
          print('Restart sln.')

          if self.it != 0 and self.nbrElmUpper != var.bNodes[1]:

            self.stagPtMvmt = 1
            print(ccolors.ANSI_BLUE, '- Stagnation point moved.',ccolors.ANSI_RESET)

          else:

            self.stagPtMvmt = 0

        else: # If we use a solution previously obtained. 

          var.u = self.uPrevious.copy()   # Use previous solution.

          print('Updating sln.')

        self.nbrElmUpper = var.bNodes[1]
        DirichletBC.airfoilBC(var)    # Reset boundary condition.

        # Run the time integration
        convergedPoints = tSolver.Run(var)
        # Output time solver status.
        if np.any(convergedPoints != 0):
            print(ccolors.ANSI_YELLOW + 'Some point(s) did not converge.', ccolors.ANSI_RESET)

        self.uPrevious=var.u.copy()   # Save solution to use it as initial condition the next iteration.

        # Compute blowing velocities and sort the boundary layer parameters.

        self.blScalars, blwVelUp, blwVelLw, blwVelWk, AFblVectors, WKblVectors, AFdeltaStarOut, WKdeltaStarOut = PostProcess().sortParam(var, self.mapMesh, cMeshDict, nFactor)
        
        self.xtr = var.xtr.copy()
        self.blVectors = AFblVectors

        AFblwVel = np.concatenate((blwVelUp, blwVelLw))

        # Group modification to ensure communication between the solvers.

        groups[0].TEnd      = [AFblVectors[1][var.bNodes[1]-1], AFblVectors[1][var.bNodes[2]-1]]
        groups[0].HEnd      = [AFblVectors[2][var.bNodes[1]-1], AFblVectors[2][var.bNodes[2]-1]]
        groups[0].CtEnd     = [AFblVectors[8][var.bNodes[1]-1], AFblVectors[8][var.bNodes[2]-1]]
        groups[0].deltaStar = AFdeltaStarOut
        groups[0].xx        = cMeshDict['locCoord'][:cMeshDict['bNodes'][2]]

        # Sort the following results in reference frame.
        groups[0].deltaStar = groups[0].deltaStar[groups[0].connectListNodes.argsort()]
        groups[0].xx        = groups[0].xx[groups[0].connectListNodes.argsort()]
        groups[0].u         = AFblwVel[groups[0].connectListElems.argsort()]
        self.data           = data[:var.bNodes[2],:]

        # Drag is measured at the end of the wake.
        self.Cd             = self.blScalars[0]
        self.Cdf            = self.blScalars[1]
        self.Cdp            = self.blScalars[2] # Cdf comes from the airfoil as there is no friction in the wake.
        groups[1].deltaStar = WKdeltaStarOut
        groups[1].xx        = cMeshDict['locCoord'][cMeshDict['bNodes'][2]:]

        # Sort the following results in reference frame.
        groups[1].deltaStar = groups[1].deltaStar[groups[1].connectListNodes.argsort()]
        groups[1].xx        = groups[1].xx[groups[1].connectListNodes.argsort()]
        groups[1].u         = blwVelWk[groups[1].connectListElems.argsort()]


        """if self.it >= 24:
          self.writeFile()
          Validation().main(1,self.wData)"""

        pass