################################################################################
#                                                                              #
#                            DARTFlo viscous module file                       #
#                          Variables: Solution and external                    #
#                             flow parameters container                        #
#                                                                              #
# Author: Paul Dechamps                                                        #
# Université de Liège                                                          #
# Date: 2021.09.10                                                             #
#                                                                              #
################################################################################


# ------------------------------------------------------------------------------
#  Imports
# ------------------------------------------------------------------------------
import numpy as np 
from matplotlib import pyplot as plt 
from fwk.coloring import ccolors

class Variables:
    """Data container for viscous solver.

    Attributes
    ----------
    
    - Ti: float [0,1]
        - User defined turbulence intensity that impacts the critical amplification ratio Ncrit of the e^N method.
    - Re: float
        - Flow Reynolds number.

    - xx: numpy array
        - Airfoil surfaces and wake mesh.

    - nN: int
        - Number of cell in the computational domain.

    - bNodes: numpy array
            - 3 elements array containing the boundary nodes of the computational domain [Stagnation point, First lower side point, First wake point].
    
    - nVar: int
          - Number of variables involved in the differential problem.

    - u: numpy array (vector of len(nVar*nN))
         - Array containing the solution (nVar variables at each of the nN points).

    - blPar: numpy array (vector of len(nBlPar*nN))
           - Boundary layer parameters.

    - extPar: numpy array (vector of len(nExtPar*nN))
            - External (inviscid) flow parameters.

    - dv: numpy array (vector of len(nN))
        - Velocity gradient on each cell.

    - xGlobal: numpy array (vector of len(nN))
             - Cell position non-dimensionalized by the airfoil chord: x/c

    - xtrF: numpy array of len(2)
          - User forced transiton on [Upper side, Lower side]
    
    - xtr: numpy array of len(2)
         - Flow laminar to turbulent transition location on [Upper side, Lower side]
        
    - Ncrit: float
            - Critical amplification ratio for the e^N method (default value= 9.) 

    """
    def __init__(self, fMeshDict, xtrF, Ti, Re):
        self.__Ti = Ti
        self.Re   = Re
        self.xx   = fMeshDict['locCoord']
        self.nN   = len(self.xx)

        self.bNodes = fMeshDict['bNodes']
        # Solution 
        self.nVar = 5
        self.u = np.zeros(self.nVar*self.nN)

        # Parameters of the boundary layers
        # [ 0 ,  1,  2,     3,     4,      5,  6,    7,  8,         9]
        # [Ret, Cd, Cf, delta, Hstar, Hstar2, Hk, Cteq, Us, deltaStar]
        self.nBlPar = 10
        self.blPar        = np.zeros(self.nBlPar*self.nN)
        self.Ret_previous = self.blPar[0::self.nVar]

        self.dueOld_dx = np.zeros(self.nN)
        self.ddeltaStarOld_dx = np.zeros(self.nN)

        # Parameters of the external (inviscid) flow
        # [ 0,    1,     2,            3,     4,    5]
        # [Me, rhoe, vtExt, deltaStarExt, xxExt ReExt]
        self.nExtPar = 6

        self.extPar                  = np.zeros(self.nExtPar*self.nN)
        self.extPar[0::self.nExtPar] = fMeshDict['Me']
        self.extPar[1::self.nExtPar] = fMeshDict['rho']
        self.extPar[2::self.nExtPar] = fMeshDict['vtExt']
        self.extPar[3::self.nExtPar] = fMeshDict['deltaStarExt']
        self.extPar[4::self.nExtPar] = fMeshDict['locCoord']
        self.dv                      = fMeshDict['dv']
        self.xGlobal                 = fMeshDict['globCoord']

        self.flowRegime = np.zeros(self.nN, dtype=int)

        self.xtrF=xtrF
        self.xtr=np.zeros(2)
        if not len(self.xtrF)==2:
            raise RuntimeError(ccolors.ANSI_RED + "Error in forced transition." + ccolors.ANSI_RESET)
        for n in range(len(self.xtrF)):
            if self.xtrF[n] is not None:
                if 0<=self.xtrF[n]<=1:
                    self.xtr[n]=self.xtrF[n]
                else:
                    raise ValueError(ccolors.ANSI_RED + "Non-physical transition location." + ccolors.ANSI_RESET)
            else:
                self.xtr[n]=1

        self.transMarkers=[self.bNodes[1]-1,self.bNodes[2]-1]

        self.Ncrit=self.turbIntensity(Ti)
        pass
    
    def turbIntensity(self,Ti):
        """Computes the critical amplification ratio Ncrit based on the input
        free-stream turbulence intensity. Ncrit=9 by default.
        """

        if Ti is not None and Ti <= 1:
            tau = 2.7 * np.tanh(self.Ti /  2.7)
            Ncrit = -8.43 - 2.4*np.log(tau / 100) # Drela 1998
        else:
            Ncrit = 9
        return Ncrit
